//******************************************************************************************************************************************************************************************//
// Public Domain                                                                                                                                                                            //
//                                                                                                                                                                                          //
// Written by Peter O. in 2014.                                                                                                                                                             //
//                                                                                                                                                                                          //
// Any copyright is dedicated to the Public Domain. http://creativecommons.org/publicdomain/zero/1.0/                                                                                       //
//                                                                                                                                                                                          //
// If you like this, you should donate to Peter O. at: http://peteroupc.github.io/                                                                                                          //
//******************************************************************************************************************************************************************************************//
using System;
using System.Collections.Generic;
using System.Text;
using Neos.IdentityServer.MultiFactor.WebAuthN.Library.Cbor.Numbers;

namespace Neos.IdentityServer.MultiFactor.WebAuthN.Library.Cbor
{
  internal sealed class OptionsParser {
    private readonly IDictionary<string, string> dict = new
    Dictionary<string, string>();

    private static string[] SplitAt(string str, string delimiter) {
      if (delimiter == null) {
        throw new ArgumentNullException(nameof(delimiter));
      }
      if (delimiter.Length == 0) {
        throw new ArgumentException("delimiter is empty.");
      }
      if (String.IsNullOrEmpty(str)) {
        return new[] { String.Empty };
      }
      var index = 0;
      var first = true;
      List<string> strings = null;
      int delimLength = delimiter.Length;
      while (true) {
        int index2 = str.IndexOf(
          delimiter,
          index,
          StringComparison.Ordinal);
        if (index2 < 0) {
          if (first) {
            var strret = new string[1];
            strret[0] = str;
            return strret;
          }
          strings = strings ?? new List<string>();
          strings.Add(str.Substring(index));
          break;
        } else {
          first = false;
          string newstr = str.Substring(index, index2 - index);
          strings = strings ?? new List<string>();
          strings.Add(newstr);
          index = index2 + delimLength;
        }
      }
      return (string[])strings.ToArray();
    }

    public OptionsParser(string options) {
      if (options == null) {
        throw new ArgumentNullException(nameof(options));
      }
      if (options.Length > 0) {
        string[] optionsArray = SplitAt(options, ";");
        foreach (string opt in optionsArray) {
          int index = opt.IndexOf('=');
          if (index < 0) {
            throw new ArgumentException("Invalid options string: " +
              options);
          }
          string key = DataUtilities.ToLowerCaseAscii(opt.Substring(0, index));
          string value = opt.Substring(index + 1);
          this.dict[key] = value;
        }
      }
    }

    public string GetLCString(string key, string defaultValue) {
      string lckey = DataUtilities.ToLowerCaseAscii(key);
      if (this.dict.ContainsKey(lckey)) {
        string lcvalue = DataUtilities.ToLowerCaseAscii(this.dict[lckey]);
        return lcvalue;
      }
      return defaultValue;
    }

    public bool GetBoolean(string key, bool defaultValue) {
      string lckey = DataUtilities.ToLowerCaseAscii(key);
      if (this.dict.ContainsKey(lckey)) {
        string lcvalue = DataUtilities.ToLowerCaseAscii(this.dict[lckey]);
        return lcvalue.Equals("1", StringComparison.Ordinal) ||
          lcvalue.Equals("yes", StringComparison.Ordinal) ||
          lcvalue.Equals("on", StringComparison.Ordinal) ||
          lcvalue.Equals("true", StringComparison.Ordinal);
      }
      return defaultValue;
    }
  }
}
